
# This file was generated by 'versioneer.py' (0.29) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2025-07-29T09:11:22+0200",
 "dirty": false,
 "error": null,
 "full-revisionid": "5672b8f5f34294613aa85748ebd61d02db199de4",
 "version": "1.9a"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
